// hacklu2013-fluxarchiv-archiv.c
//
// make hacklu2013-fluxarchiv-archiv CFLAGS="--std=c99 -lcrypto"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include <openssl/sha.h>
#include <openssl/rc4.h>
#include <openssl/md5.h>

const char*         MAGIC_VALUE     = "FluXArChiV13";
const unsigned int  MAGIC_VALUE_SZ  = 0x0C;

const unsigned int  PASSWD_CHECK_OFFSET = 0x0C;

const unsigned int  FILENAME_SZ = 0x60;

const unsigned int  BLOCK_SZ = 0x408;

char hash_of_password[SHA_DIGEST_LENGTH];

char buffer[0x410];

const char*         FLUXLIST_MAGIC      = "FluXL1sT";
const unsigned int  FLUXLIST_MAGIC_SZ   = 8;
const unsigned int  FLUXLIST_OFFSET     = 0x20;

/********** THIS FUNCTION IS OK **********/
void sanitizeFilename(char* dest, int n) { // decrypt_data
    RC4_KEY key;
    char* src;

    src = calloc(n, 1);
    RC4_set_key(&key, SHA_DIGEST_LENGTH, hash_of_password);
    RC4(&key, n, dest, src);
    memcpy(dest, src, n);
    free(src);
}

/********** THIS FUNCTION IS OK **********/
uint64_t read_int(FILE* stream) {
    char ptr[8];
    uint64_t res = 0;

    fread(ptr, 8, 1, stream);
    sanitizeFilename(ptr, 8);
    for (int counter = 0; counter < 8; ++counter)
        res = (res << 8) | (uint8_t) ptr[7 - counter];
    return res;
}

void createHashOfPassword(const char* src, char* dest) { // basename
    char* ret;

    ret = strrchr(src, '/');
    if (ret != NULL) {
        src = ret + 1;
    }
    strncpy(dest, src, FILENAME_SZ - 1);
    dest[FILENAME_SZ - 1] = '\0';
}

/********** THIS FUNCTION IS OK **********/
int encryptDecryptData(FILE* stream) { // check_magic_num_archive
    char ptr[MAGIC_VALUE_SZ];

    fseek(stream, 0, 0);
    fread(ptr, MAGIC_VALUE_SZ, 1, stream);
    return memcmp(MAGIC_VALUE, ptr, MAGIC_VALUE_SZ) == 0;
}

/********** THIS FUNCTION IS OK **********/
void checkHashOfPassword(char* s) { // calculate_hash_of_password
    SHA_CTX ctx;

    SHA1_Init(&ctx);
    SHA1_Update(&ctx, s, strlen(s));
    SHA1_Final(hash_of_password, &ctx);
}

/********** THIS FUNCTION IS OK **********/
int verifyArchiv(FILE* stream) { // check_passwd
    int idx[SHA_DIGEST_LENGTH] = {0, 7, 14, 1, 8, 15, 2, 9, 16, 3, 10,
                                  17, 4, 11, 18, 5, 12, 19, 6, 13};
    unsigned int counter;
    char ptr[SHA_DIGEST_LENGTH];
    char tmp[SHA_DIGEST_LENGTH];
    char tmp2[SHA_DIGEST_LENGTH];
    SHA_CTX ctx;

    fseek(stream, PASSWD_CHECK_OFFSET, SEEK_SET);
    fread(ptr, SHA_DIGEST_LENGTH, 1, stream);
    for (counter = 0; counter < SHA_DIGEST_LENGTH; ++counter) {
        tmp[counter] = hash_of_password[idx[counter]];
    }

    SHA1_Init(&ctx);
    SHA1_Update(&ctx, tmp, SHA_DIGEST_LENGTH);
    SHA1_Final(tmp2, &ctx);

    return memcmp(ptr, tmp2, SHA_DIGEST_LENGTH) == 0;
}

void clear_data(FILE* stream, unsigned int size) {
    char* buffer = NULL;

    buffer = calloc(size, 1);
    sanitizeFilename(buffer, size);
    fwrite(buffer, size, 1, stream);
    free(buffer);
}

void listAllFilesInArchiv(FILE* stream, unsigned int off) { // delete_file
    char ptr[0x8];
    uint64_t counter;
    uint64_t curpos;
    uint64_t nbblocks;
    uint64_t nextblock;

    fseek(stream, off, SEEK_SET);
    nextblock = read_int(stream);
    nbblocks = read_int(stream);
    fseek(stream, off, SEEK_SET);
    clear_data(stream, 8);
    clear_data(stream, 8);
    clear_data(stream, MD5_DIGEST_LENGTH);
    clear_data(stream, FILENAME_SZ);
    for (counter = 0; counter < nbblocks; ++counter) {
        fseek(stream, nextblock * 1040 + 0x20, SEEK_SET);
        curpos = ftell(stream);
        nextblock = read_int(stream);
        fseek(stream, curpos, SEEK_SET);
        clear_data(stream, 8);
    }
}

/********** THIS FUNCTION IS OK **********/
void deleteFileFromArchive(FILE* stream, FILE* s, char* hash, unsigned int off) {
    MD5_CTX md5Ctx;
    uint64_t counter;
    uint64_t curpos;
    uint64_t nextoff;
    uint64_t nbblocks;
    uint64_t size;
    uint64_t n;

    MD5_Init(&md5Ctx);
    fseek(stream, off, SEEK_SET);
    nextoff = read_int(stream);
    nbblocks = read_int(stream);
    for (counter = 0; counter < nbblocks; ++counter) {
        printf("offset = %x\n", nextoff * 1040 + 0x20);
        fseek(stream, nextoff * 1040 + 0x20, SEEK_SET);
        curpos = ftell(stream);
        nextoff = read_int(stream);
        if (counter != nbblocks - 1) {
            fread(buffer, 1, BLOCK_SZ, stream);
            sanitizeFilename(buffer, BLOCK_SZ);
            fwrite(buffer, 1, BLOCK_SZ, s);
            MD5_Update(&md5Ctx, buffer, BLOCK_SZ);
        } else {
            n = nextoff + BLOCK_SZ + (((-nbblocks << 7) - nbblocks) << 3);
            fread(buffer, 1, n, stream);
            sanitizeFilename(buffer, n);
            fwrite(buffer, 1, n, s);
            MD5_Update(&md5Ctx, buffer, n);
        }
    }
    MD5_Final(hash, &md5Ctx);
}

/********** THIS FUNCTION IS OK **********/
int extractFileFromArchiv(FILE* stream, const char* s2) { // search_offset
    char ptr[0x100];
    uint64_t tmp2, tmp3 = 0, res = 0;

    fseek(stream, FLUXLIST_OFFSET, SEEK_SET);
    do {
        fread(ptr, FLUXLIST_MAGIC_SZ, 1, stream);
        sanitizeFilename(ptr, FLUXLIST_MAGIC_SZ);
        if (memcmp(FLUXLIST_MAGIC, ptr, FLUXLIST_MAGIC_SZ) != 0) {
            puts("Could not locate file entry list => file corrupted.");
            exit(1);
        }
        tmp2 = read_int(stream);
        for (int counter2 = 0; counter2 < 8; ++counter2) {
            fseek(stream, tmp3 * 1040 + 0x50 + counter2 * 0x80, SEEK_SET);
            fread(ptr, FILENAME_SZ, 1, stream);
            sanitizeFilename(ptr, FILENAME_SZ);
            if (strcmp(ptr, s2) == 0) {
                res = tmp3 * 1040 + 0x30 + counter2 * 0x80;
                break;
            }
        }
        if (tmp2 == 0 || res != 0)
            break;
        fseek(stream, tmp2 * 1040 + 0x20, SEEK_SET);
        tmp3 = tmp2;
    } while (1);
    return res;
}

#if 0
int createFileEntry(FILE* stream) {
    long curpos, max, result = 0;
    int a;
    char ptr[8];
    uint64_t counter;
    uint64_t size;

    // Fetch the size of the stream.
    fseek(stream, 0, SEEK_END);
    size = ftell(stream);
    max = (((curpos - 0x20) * 0x0FC0FC0FC0FC0FC1) >> 6) - (curpos >> 0x3F);
    for (counter = 0; counter < max; ++counter) {
        fseek(stream, counter * 1040 + 0x20, SEEK_SET);
        a = read_int(stream);
        if (a == 0)
            break;
    }
    result = counter;
    if (result == 0)
        result = max;
    return result;
}

void searchFileInArchiv(FILE* s, FILE* stream, int arg3, int arg4) {
    uint64_t size = 0, wtf1, wtf2, wtf3;
    int off;
    uint64_t counter = 0;                        // counter for blocks.
    int percent_counter = 0xA;              // Counter for percentage displaying.
    MD5_CTX md5Ctx;                         // MD5 context for generation.
    char ptr[0x100];

    MD5_Init(&md5Ctx);
    // Fetch the size of the stream.
    fseek(stream, 0, SEEK_END);
    size = ftell(stream);
    // Go back to the beginning of the file.
    fseek(stream, 0, SEEK_SET);
    // WTF: find out wtf this is....
    wtf1 = ((size * 0xfe03f80fe03f81) >> 2) - (size >> 0x3f) + 1;
    // Begin progress display.
    puts("\nProgress:");
    // WTF2: find out wtf this is too...
    wtf2 = wtf1 - (wtf1 * 0x1c71c71c71c71c72) >> 0x3f;
    // Display percent 0.
    printf("0%%");
    fflush(stdout);
    for (counter = 0; counter < wtf1; /* done in the loop */) {
        fseek(s, arg4 * 1040 + 0x20, SEEK_SET);
        off = ftell(s);
        fwrite("keepfree", 8, 1, s);
        fread(buffer, 1, BLOCK_SZ, stream);
        counter += 1;
        if (counter == wtf1) {
            wtf3 = size + BLOCK_SZ - (wtf1 << 7) - (wtf1 << 3);
            MD5_Update(&md5Ctx, buffer, wtf3);
        } else {
            MD5_Update(&md5Ctx, buffer, BLOCK_SZ);
        }
        sanitizeFilename(buffer, BLOCK_SZ);
        fwrite(buffer, BLOCK_SZ, 1, s);
        if (wtf2 != 0) {
            printf(" ... %d%%", percent_counter);
            fflush(stdout);
            percent_counter += 0xA;
        } else if ((counter >> 0x3f) / wtf2) {
            printf(" ... %d%%", percent_counter);
            fflush(stdout);
            percent_counter += 0xA;
        }
        if (counter == wtf1 - 1) {
            arg4 = size;
        } else {
            arg4 = createFileEntry(s);
        }
        fseek(s, off, SEEK_SET);
        for (int counter2 = 0; counter2 < 7; counter2) {
            ptr[counter2] = arg4 >> (counter2 << 3);
        }
        sanitizeFilename(ptr, 8);
        fwrite(ptr, 8, 1, s);
    }
}
#endif

/********** THIS FUNCTION IS OK **********/
void writeFileToArchiv(FILE* stream) { // list_files_in_archive
    char ptr[0x100];
    uint64_t counter2;
    uint64_t var_10 = 0;
    uint64_t var_20;
    uint64_t var_8;

    puts("Filename:\t\t\t\t\t\t\tSize in archiv:");
    puts("--------------------------------------------------------------");
    fseek(stream, FLUXLIST_OFFSET, SEEK_SET);
    do {
        // Check if FLUXLIST_MAGIC can be read correctly.
        fread(ptr, FLUXLIST_MAGIC_SZ, 1, stream);
        sanitizeFilename(ptr, FLUXLIST_MAGIC_SZ);
        if (memcmp(FLUXLIST_MAGIC, ptr, FLUXLIST_MAGIC_SZ) != 0) {
            puts("Could not locate file entry list => file corrupted.");
            exit(1);
        }
        var_8 = read_int(stream);
        for (counter2 = 0; counter2 <= 7; ++counter2) {
            fseek(stream, var_10 * 1040 + 0x38 + counter2 * 0x80, SEEK_SET);
            var_20 = read_int(stream);
            if (var_20 != 0) {
                fseek(stream, var_10 * 1040 + 0x50 + counter2 * 0x80, SEEK_SET);
                fread(ptr, FILENAME_SZ, 1, stream);
                sanitizeFilename(ptr, FILENAME_SZ);
                printf("%-50s%20ld kB\n", ptr, (var_20 * 1040) / 1000);
            }
        }
        if (var_8 == 0)
            return;
        fseek(stream, var_8 * 1040 + 0x20, SEEK_SET);
        var_10 = var_8;
    } while (1);
}

#if 0
int findNextFreeChunk(FILE* s, FILE* stream, ...) {
    uint64_t stream_size;

    fseek(stream, 0, SEEK_END);
    stream_size = ftell(stream);
    fseek(stream, 0, SEEK_SET);
    createFileEntry(s);
}
#endif

int main(int argc, char** argv) {
    char hash[MD5_DIGEST_LENGTH];
    char hash2[MD5_DIGEST_LENGTH];
    uint64_t offset = 0;
    int ret = 0;
    FILE* s1 = NULL;
    FILE* s2 = NULL;

    puts("Welcoming message.");
    if (argc == 4) {
        if (strcmp("-l", argv[1]) == 0) {
            checkHashOfPassword(argv[3]);
            if (access(argv[2], R_OK) == -1)
                goto cannot_open_archiv;
            s1 = fopen(argv[2], "r");
            if (!encryptDecryptData(s1))
                goto archiv_invalid;
            if (!verifyArchiv(s1))
                goto password_incorrect;
            writeFileToArchiv(s1);
        }
    } else if (argc == 5) {
        if (strcmp("-x", argv[1]) == 0) {
            checkHashOfPassword(argv[3]);
            if (access(argv[2], R_OK) == -1)
                goto cannot_open_archiv;
            s1 = fopen(argv[2], "r");
            if (!encryptDecryptData(s1))
                goto archiv_invalid;
            if (!verifyArchiv(s1))
                goto password_incorrect;
            offset = extractFileFromArchiv(s1, argv[4]);
            if (offset == 0)
                goto file_not_found;
            if (access(argv[4], F_OK) != -1) {
                puts("File with same filename already existed.");
                return 1;
            }
            s2 = fopen(argv[4], "w");
            deleteFileFromArchive(s1, s2, hash, offset);
            fseek(s1, offset + 0x10, SEEK_SET);
            fread(hash2, MD5_DIGEST_LENGTH, 1, s1);
            sanitizeFilename(hash2, MD5_DIGEST_LENGTH);
            if (memcmp(hash, hash2, MD5_DIGEST_LENGTH) != 0)
                goto md5_hash_doesnotmatch;
            printf("File %s successfully extracted from archiv.\n", argv[2]);
        }
    }

end:
    if (s1 != NULL)
        fclose(s1);
    if (s2 != NULL)
        fclose(s2);
    return ret;

file_not_found:
    puts("File was not found in archiv.");
    ret = 1; goto end;

cannot_open_archiv:
    puts("Archiv cannot be opened with read access...");
    ret = 1; goto end;

archiv_invalid:
    puts("Given archiv file is not valid.");
    ret = 1; goto end;

password_incorrect:
    puts("Given password is not correct.");
    ret = 1; goto end;

md5_hash_doesnotmatch:
    printf("MD5 hash of file %s does not match => file is corrupt :(\n", argv[2]);
    ret = 1; goto end;
}
